/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// gl_anim.c -- handle warping sky/water animation, fog and bloom (2D lighting post process effect) routines

#include "quakedef.h"

gltexture_t		*skyboxtextures[6];
gltexture_t		*solidskytexture, *alphaskytexture;

cvar_t	r_fastsky = {"r_fastsky","0"};
cvar_t	r_skyquality = {"r_skyquality","12"};
cvar_t	r_skyalpha = {"r_skyalpha","1"};
cvar_t	r_skyfog = {"r_skyfog","0.5"};
cvar_t	r_oldsky = {"r_oldsky", "0"};

cvar_t	r_waterquality = {"r_waterquality", "12"};

cvar_t	gl_fogenable = {"gl_fogenable", "0"};
cvar_t	gl_fogdensity = {"gl_fogdensity", "0"};
cvar_t	gl_fogred = {"gl_fogred","0.3"};
cvar_t	gl_foggreen = {"gl_foggreen","0.3"};
cvar_t	gl_fogblue = {"gl_fogblue","0.3"};

// speed up sin calculations - Ed
float turbsin[] =
{
	#include "gl_warp_sin.h"
};

qboolean oldsky;

int		sky_width[6], sky_height[6];
char	skybox_name[MAX_QPATH] = ""; // name of current skybox, or "" if no skybox
char	*suf[6] = {"rt", "bk", "lf", "ft", "up", "dn"}; // for skybox

// skybox
vec3_t	skyclip[6] = {
	{1,1,0},
	{1,-1,0},
	{0,-1,1},
	{0,1,1},
	{1,0,1},
	{-1,0,1}
};

int	st_to_vec[6][3] =
{
	{3,-1,2},
	{-3,1,2},
	{1,3,2},
	{-1,-3,2},
 	{-2,-1,3},		// 0 degrees yaw, look straight up
 	{2,-1,-3}		// look straight down
};

int	vec_to_st[6][3] =
{
	{-2,3,1},
	{2,3,-1},
	{1,3,2},
	{-1,3,-2},
	{-2,-1,3},
	{-2,1,-3}
};

float	skymins[2][6], skymaxs[2][6];
int		skytexorder[6] = {0,2,1,3,4,5}; // for skybox

#define	MAX_CLIP_VERTS 64

float	skyflatcolor[3];

float fog_density;
float fog_red;
float fog_green;
float fog_blue;

//==================================================================================================================

/*
=================================================================

  PCX Loading

=================================================================
*/

#pragma pack(1)
typedef struct
{
    char	    manufacturer;
    char	    version;
    char	    encoding;
    char	    bits_per_pixel;
    unsigned short  xmin,ymin,xmax,ymax;
    unsigned short  hres,vres;
    byte   palette[48];
    char	    reserved;
    char	    color_planes;
    unsigned short  bytes_per_line;
    unsigned short  palette_type;
    char	    filler[58];
} pcx_t;
#pragma pack()

static unsigned short SwapShort (unsigned short *SVal)
{
	byte *Buf = (byte *)SVal;

	return (unsigned short)(Buf[0] + Buf[1] * 256);
}

static void ChkBounds (char *Function, int Val, int MaxVal, char *FileName)
{
	if (Val > MaxVal)
		Sys_Error ("%s: invalid buffer access (%d, max = %d) in %s", Function, Val, MaxVal, FileName);
}

/*
============
LoadPCX
============
*/
static byte *LoadPCX (char *name, int *width, int *height, qboolean alphablend)
{
	pcx_t	    *pcx_header;
	FILE	    *f;
	byte	    *palette, *palid;
	byte	    *pix;
	byte	    *pcx_rgb, *buf;
	int	    x, y;
	int	    dataByte, runLength;
	int	    count, rows, columns;
	char	    filename[MAX_OSPATH];
	qboolean    blend;

	sprintf (filename, "%s.pcx", name);
	COM_FOpenFile (filename, &f);

	if (!f)
		return NULL;

	if (com_filesize < sizeof(pcx_t))
		Sys_Error ("LoadPCX: can't read header in %s", filename);

	pcx_header = malloc (com_filesize);

	if (!pcx_header)
		Sys_Error ("LoadPCX1: error allocating %d bytes for %s", com_filesize, filename);

	if ((int)fread(pcx_header, 1, com_filesize, f) != com_filesize)
		Sys_Error ("LoadPCX: error reading %s", filename);

	fclose (f);

	// get palette
	palette = (byte *)pcx_header + com_filesize - 768;
	palid = palette - 1;
	
	if (*palid != 12)
	{
		// Palette id invalid; assume missing
		Con_SafePrintf ("LoadPCX: palette id invalid (%d, should be 12) in %s\n", *palid, name);
		++palid;
	}

//
// parse the PCX file
//
	if (pcx_header->manufacturer != 0x0a
		|| pcx_header->version != 5
		|| pcx_header->encoding != 1
		|| pcx_header->bits_per_pixel * pcx_header->color_planes != 8)
		Sys_Error ("LoadPCX: bad format in %s (mf=%d, ver=%d, enc=%d, bpp=%d, pal=%d)", filename, pcx_header->manufacturer,
			   pcx_header->version, pcx_header->encoding, pcx_header->bits_per_pixel * pcx_header->color_planes, *palid);

	pcx_header->xmin = SwapShort (&pcx_header->xmin);
	pcx_header->ymin = SwapShort (&pcx_header->ymin);
	pcx_header->xmax = SwapShort (&pcx_header->xmax);
	pcx_header->ymax = SwapShort (&pcx_header->ymax);
	pcx_header->xmin = SwapShort (&pcx_header->xmin);

	columns = pcx_header->xmax - pcx_header->xmin + 1;
	rows = pcx_header->ymax - pcx_header->ymin + 1;
	count = columns * rows;
	count += 8; // If I don't add a few pixels here, validation LoadPCX3 below will sometimes fire (?)

	pcx_rgb = malloc (count * 4);

	if (!pcx_rgb)
		Sys_Error ("LoadPCX2: error allocating %d bytes for %s", count * 4, filename);

	blend = alphablend;
	buf = (byte *)(pcx_header + 1);

	for (y=0 ; y<rows ; y++)
	{
		pix = pcx_rgb + 4*y*columns;
		for (x=0 ; x<columns ; )
		{
			ChkBounds ("LoadPCX1", buf - (byte *)pcx_header, palid - 1 - (byte *)pcx_header, filename);
			dataByte = *buf++;

			if ((dataByte & 0xC0) == 0xC0)
			{
				runLength = dataByte & 0x3F;
				ChkBounds ("LoadPCX2", buf - (byte *)pcx_header, palid - 1 - (byte *)pcx_header, filename);
				dataByte = *buf++;
			}
			else
				runLength = 1;

			ChkBounds ("LoadPCX3", pix - pcx_rgb + runLength * 4, count * 4, filename);

			while(runLength-- > 0)
			{
				pix[0] = palette[dataByte*3];
				pix[1] = palette[dataByte*3+1];
				pix[2] = palette[dataByte*3+2];
				pix[3] = blend ? (pix[0] + pix[1] + pix[2]) / 3 : 255; // Better way?
				pix += 4;
				x++;
			}
		}
	}

	*width = columns;
	*height = rows;

	free (pcx_header);

	return pcx_rgb;
}

/*
=========================================================

TARGA LOADING

=========================================================
*/

#pragma pack(1)
typedef struct _TargaHeader {
	byte 	id_length, colormap_type, image_type;
	unsigned short	colormap_index, colormap_length;
	byte	colormap_size;
	unsigned short	x_origin, y_origin, width, height;
	byte	pixel_size, attributes;
} TargaHeader;
#pragma pack()

/*
=============
LoadTGA
=============
*/
static byte *LoadTGA (char *name, int *width, int *height, qboolean alphablend)
{
	TargaHeader	*targa_header;
	FILE		*f;
	int			columns, rows, numPixels;
	byte		*pixbuf, *buf;
	byte		*targa_rgba;
	int			row, realrow, column;
	qboolean	upside_down, alpha, blend;
	char		filename[MAX_OSPATH];

	sprintf (filename, "%s.tga", name);
	COM_FOpenFile (filename, &f);

	if (!f)
		return NULL;

	if (com_filesize < sizeof(TargaHeader))
		Sys_Error ("LoadTGA: can't read header in %s", filename);

	targa_header = malloc (com_filesize);

	if (!targa_header)
		Sys_Error ("LoadTGA1: error allocating %d bytes for %s", com_filesize, filename);

	if ((int)fread(targa_header, 1, com_filesize, f) != com_filesize)
		Sys_Error ("LoadTGA: error reading %s", filename);

	fclose (f);

	targa_header->width = SwapShort (&targa_header->width);
	targa_header->height = SwapShort (&targa_header->height);

	if (targa_header->image_type!=2
		&& targa_header->image_type!=10)
		Sys_Error ("LoadTGA: %s: only type 2 and 10 targa RGB images supported, not %d", filename, targa_header->image_type);

	if (targa_header->colormap_type !=0
		|| (targa_header->pixel_size!=32 && targa_header->pixel_size!=24))
		Sys_Error ("LoadTGA: %s: only 24 and 32 bit images supported (no colormaps, type=%d, bpp=%d)", filename, targa_header->colormap_type, targa_header->pixel_size);

	columns = targa_header->width;
	rows = targa_header->height;
	numPixels = columns * rows;
	upside_down = !(targa_header->attributes & 0x20); // true => picture is stored bottom to top

	targa_rgba = malloc (numPixels * 4);

	if (!targa_rgba)
		Sys_Error ("LoadTGA2: error allocating %d bytes for %s", numPixels*4, filename);

	blend = alphablend;
	alpha = targa_header->pixel_size == 32;

	buf = (byte *)(targa_header + 1);

	if (targa_header->id_length != 0)
		buf += targa_header->id_length; // skip TARGA image comment

	if (targa_header->image_type==2) {  // Uncompressed, RGB images
		for(row=rows-1; row>=0; row--) {
			realrow = upside_down ? row : rows - 1 - row;
			pixbuf = targa_rgba + realrow*columns*4;

			ChkBounds ("LoadTGA1", buf - (byte *)targa_header + columns * targa_header->pixel_size / 8, com_filesize, filename);
			ChkBounds ("LoadTGA2", pixbuf - targa_rgba + columns * 4, numPixels * 4, filename);

			for(column=0; column<columns; column++) {
				byte red,green,blue;
				blue = *buf++;
				green = *buf++;
				red = *buf++;
				*pixbuf++ = red;
				*pixbuf++ = green;
				*pixbuf++ = blue;
				*pixbuf++ = alpha ? *buf++ : (blend ? (red + green + blue) / 3 : 255); // Better way?;
			}
		}
	}
	else if (targa_header->image_type==10) {   // Runlength encoded RGB images
		byte red,green,blue,alphabyte,packetHeader,packetSize,j;
		for(row=rows-1; row>=0; row--) {
			realrow = upside_down ? row : rows - 1 - row;
			pixbuf = targa_rgba + realrow*columns*4;
			for(column=0; column<columns; ) {
				packetHeader = *buf++;
				packetSize = 1 + (packetHeader & 0x7f);

				ChkBounds ("LoadTGA3", pixbuf - targa_rgba + packetSize * 4, numPixels * 4, filename);

				if (packetHeader & 0x80) {        // run-length packet
					ChkBounds ("LoadTGA4", buf - (byte *)targa_header + targa_header->pixel_size / 8, com_filesize, filename);
					blue = *buf++;
					green = *buf++;
					red = *buf++;
					alphabyte = alpha ? *buf++ : (blend ? (red + green + blue) / 3 : 255); // Better way?;

					for(j=0;j<packetSize;j++) {
						*pixbuf++ = red;
						*pixbuf++ = green;
						*pixbuf++ = blue;
						*pixbuf++ = alphabyte;
						column++;
						if (column==columns) { // run spans across rows
							column=0;
							if (row>0)
								row--;
							else
								goto breakOut;
							realrow = upside_down ? row : rows - 1 - row;
							pixbuf = targa_rgba + realrow*columns*4;
						}
					}
				}
				else {                            // non run-length packet
					ChkBounds ("LoadTGA5", buf - (byte *)targa_header + packetSize * targa_header->pixel_size / 8, com_filesize, filename);
					for(j=0;j<packetSize;j++) {
						blue = *buf++;
						green = *buf++;
						red = *buf++;
						*pixbuf++ = red;
						*pixbuf++ = green;
						*pixbuf++ = blue;
						*pixbuf++ = alpha ? *buf++ : (blend ? (red + green + blue) / 3 : 255); // Better way?;
						column++;
						if (column==columns) { // pixel packet run spans across rows
							column=0;
							if (row>0)
								row--;
							else
								goto breakOut;
							realrow = upside_down ? row : rows - 1 - row;
							pixbuf = targa_rgba + realrow*columns*4;
						}
					}
				}
			}
			breakOut:;
		}
	}

	*width = targa_header->width;
	*height = targa_header->height;

	free (targa_header);

	return targa_rgba;
}

/*
============
R_LoadImage
============
*/
byte *R_LoadImage (char *name, int *width, int *height, qboolean alphablend)
{
	byte *data = LoadTGA (name, width, height, alphablend);

	if (data)
		return data;

	return LoadPCX (name, width, height, alphablend);
}


//==================================================================================================================


/*
==============================================================================

	RENDER-TO-FRAMEBUFFER WATER (adapted from fitzquake)

==============================================================================
*/

/*
=============
R_UpdateWarpTextures

each frame, update warping textures
=============
*/
void R_UpdateWarpTextures (void)
{
	texture_t *tx;
	int i;
	float x, y, x2, warptess;

	warptess = 128.0/CLAMP (4.0, floor(r_waterquality.value), 64.0);

	for (i=0; i<cl.worldmodel->numtextures; i++)
	{
		if (!(tx = cl.worldmodel->textures[i]))
			continue;

		if (tx->name[0] != '*')
			continue;

		if (!tx->update_warp)
			continue;

		//render warp
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity ();
		glOrtho (0, 128, 0, 128, -99999, 99999);
		glViewport (glx, gly + glheight - gl_warp_image_size, gl_warp_image_size, gl_warp_image_size);
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity ();

		GL_Bind (tx->gltexture);
		for (x=0.0; x<128.0; x=x2)
		{
			x2 = x + warptess;
			glBegin (GL_TRIANGLE_STRIP);
			for (y=0.0; y<128.01; y+=warptess) // .01 for rounding errors
			{
				glTexCoord2f (WARPCALC(x,y), WARPCALC(y,x));
				glVertex2f (x,y);
				glTexCoord2f (WARPCALC(x2,y), WARPCALC(y,x2));
				glVertex2f (x2,y);
			}
			glEnd();
		}

		//copy to texture
		GL_Bind (tx->warpimage);
		glCopyTexSubImage2D (GL_TEXTURE_2D, 0, 0, 0, glx, gly + glheight - gl_warp_image_size, gl_warp_image_size, gl_warp_image_size);

		tx->update_warp = false;
	}

	//if warp render went down into sbar territory, we need to be sure to refresh it next frame
	if (gl_warp_image_size + sb_lines > glheight)
		Sbar_Changed ();
}


/*
==============================================================================

	FOG DRAW ROUTINES

==============================================================================
*/

/*
=============
R_FogUpdate

update internal variables
=============
*/
void R_FogUpdate (float density, float red, float green, float blue)
{
	Cvar_SetValue ("gl_fogenable", density ? 1 : 0);
	Cvar_SetValue ("gl_fogdensity", density);
	Cvar_SetValue ("gl_fogred", red);
	Cvar_SetValue ("gl_foggreen", green);
	Cvar_SetValue ("gl_fogblue", blue);
}

/*
=============
R_FogParseServerMessage

handle an 'svc_fog' message from server
=============
*/
void R_FogParseServerMessage (void)
{
	float density, red, green, blue;

	density = MSG_ReadFloat(net_message);
	red = MSG_ReadByte(net_message) / 255.0;
	green = MSG_ReadByte(net_message) / 255.0;
	blue = MSG_ReadByte(net_message) / 255.0;

	R_FogUpdate (density, red, green, blue);
}

/*
=============
R_Fog_f

handle the 'fog' console command
=============
*/
void R_Fog_f (void)
{
	switch (Cmd_Argc())
	{
		default:
		case 1:
			Con_Printf ("usage:\n");
			Con_Printf ("   fog <density>\n");
			Con_Printf ("   fog <density> <rgb>\n");
			Con_Printf ("   fog <red> <green> <blue>\n");
			Con_Printf ("   fog <density> <red> <green> <blue>\n");
			Con_Printf("current values:\n");
			Con_Printf("   fog is %sabled\n", gl_fogenable.value ? "en" : "dis");
			Con_Printf("   density is %f\n", gl_fogdensity.value);
			Con_Printf("   red   is %f\n", gl_fogred.value);
			Con_Printf("   green is %f\n", gl_foggreen.value);
			Con_Printf("   blue  is %f\n", gl_fogblue.value);
			break;
		case 2:
			R_FogUpdate(max(0.0, atof(Cmd_Argv(1))), gl_fogred.value, gl_foggreen.value, gl_fogblue.value);
			break;
		case 3:
			R_FogUpdate(max(0.0, atof(Cmd_Argv(1))), CLAMP(0.0, atof(Cmd_Argv(2)), 1.0), CLAMP(0.0, atof(Cmd_Argv(2)), 1.0), CLAMP(0.0, atof(Cmd_Argv(2)), 1.0));
			break;
		case 4:
			R_FogUpdate(gl_fogdensity.value, CLAMP(0.0, atof(Cmd_Argv(1)), 1.0), CLAMP(0.0, atof(Cmd_Argv(2)), 1.0), CLAMP(0.0, atof(Cmd_Argv(3)), 1.0));
			break;
		case 5:
			R_FogUpdate(max(0.0, atof(Cmd_Argv(1))), CLAMP(0.0, atof(Cmd_Argv(2)), 1.0), CLAMP(0.0, atof(Cmd_Argv(3)), 1.0), CLAMP(0.0, atof(Cmd_Argv(4)), 1.0));
			break;
	}
}

/*
=============
R_FogGetColor

calculates fog color
=============
*/
float *R_FogGetColor (void)
{
	static float c[4];
	int i;

	c[0] = gl_fogred.value;
	c[1] = gl_foggreen.value;
	c[2] = gl_fogblue.value;
	c[3] = 1.0;

	//find closest 24-bit RGB value, so solid-colored sky can match the fog perfectly
	for (i=0;i<3;i++)
		c[i] = (float)(Q_rint(c[i] * 255)) / 255.0f;

	return c;
}

/*
=============
R_FogGetDensity

returns current density of fog
=============
*/
float R_FogGetDensity (void)
{
	if (gl_fogenable.value)
		return gl_fogdensity.value;
	else
		return 0;
}

/*
=============
R_FogSetupFrame

called at the beginning of each frame
=============
*/
void R_FogSetupFrame (void)
{
	glFogfv(GL_FOG_COLOR, R_FogGetColor());
	glFogf(GL_FOG_DENSITY, R_FogGetDensity() / 64.0);
}

/*
=============
R_FogEnableGFog

called before drawing stuff that should be fogged
=============
*/
void R_FogEnableGFog (void)
{
	if (R_FogGetDensity() > 0)
		glEnable(GL_FOG);
}

/*
=============
R_FogDisableGFog

called after drawing stuff that should be fogged
=============
*/
void R_FogDisableGFog (void)
{
	if (R_FogGetDensity() > 0)
		glDisable(GL_FOG);
}

/*
=============
R_FogStartAdditive

called before drawing stuff that is additive blended -- sets fog color to black
=============
*/
void R_FogStartAdditive (void)
{
	vec3_t color = {0,0,0};

	if (R_FogGetDensity() > 0)
		glFogfv(GL_FOG_COLOR, color);
}

/*
=============
R_FogStopAdditive

called after drawing stuff that is additive blended -- restores fog color
=============
*/
void R_FogStopAdditive (void)
{
	if (R_FogGetDensity() > 0)
		glFogfv(GL_FOG_COLOR, R_FogGetColor());
}


/*
==============================================================================

	SKY DRAW ROUTINES

==============================================================================
*/

/*
=============================================================

	RENDER SKYBOX

=============================================================
*/
/*
==================
R_LoadSkyBox
==================
*/
void R_LoadSkyBox (char *skybox)
{
	int		i, j;
	char	name[MAX_QPATH];
	byte	*data[6];

	if (!strcmp(skybox, skybox_name)) //no change
	{
		oldsky = (skybox_name[0] == 0);
		return;
	}

	//purge old textures
	for (i = 0; i < 6; i++)
	{
		if (skyboxtextures[i] && skyboxtextures[i] != notexture)
			GL_FreeTexture (skyboxtextures[i]);
		skyboxtextures[i] = NULL;
	}

	// First try gfx/env and then env
	for (j = 0 ; j < 2 ; j++)
	{
		//load textures
		for (i = 0 ; i < 6 ; i++)
		{
			sprintf (name, "%senv/%s%s", j == 0 ? "gfx/" : "", skybox, suf[i]);
			data[i] = R_LoadImage (name, &sky_width[i], &sky_height[i], false);
			if (!data[i])
				break;
		}

		if (i == 6)
			break;
	}

	if (i == 6)
	{
		for (j = 0; j < 6; ++j)
		{
			sprintf (name, "gfx/env/%s%s", skybox, suf[j]);
			skyboxtextures[j] = GL_LoadTexture (cl.worldmodel, name, sky_width[j], sky_height[j], SRC_RGBA, data[j], TEXPREF_NONE);
		}
	}
	else
	{
		Con_Printf ("Couldn't load skybox %s\n", skybox);
		for (j = 0; j < 6; ++j)
			skyboxtextures[j] = notexture;
	}

	for (j = 0; j < i; ++j)
		free (data[j]);

	if (i == 6)
	{
		strcpy (skybox_name, skybox);
		Con_DPrintf ("skybox set to %s\n", skybox);
	}
	else
	{
		for (j = 0; j < 6; j++)
		{
			if (skyboxtextures[j] && skyboxtextures[j] != notexture)
				GL_FreeTexture (skyboxtextures[j]);
			skyboxtextures[j] = NULL;
		}
	}

	// Enable/disable skybox
	oldsky = (skybox_name[0] == 0);
}

/*
=================
R_InitSkyBox

called when quake initializes
also handle fog variables
=================
*/
void R_InitSkyBox (void)
{
	int i;

	//clear skyboxtextures pointers
	for (i=0; i<6; i++)
		skyboxtextures[i] = NULL;

	//set up global fog
	glFogi(GL_FOG_MODE, GL_EXP2);
}

/*
=================
R_SkyNewMap

also handle fog variables
=================
*/
void R_SkyNewMap (void)
{
	char  key[MAX_KEY], value[MAX_VALUE];
	char  *data;
	int i;

	// initially no skybox
	oldsky = true;
	strcpy (skybox_name, "");
	for (i = 0; i < 6; i++)
		skyboxtextures[i] = NULL;

	//initially no fog
	fog_density = 0;
	fog_red = 0.3;
	fog_green = 0.3;
	fog_blue = 0.3;

	Cvar_SetValue ("gl_fogenable", 0);
	Cvar_SetValue ("gl_fogdensity", fog_density);
	Cvar_SetValue ("gl_fogred", fog_red);
	Cvar_SetValue ("gl_foggreen", fog_green);
	Cvar_SetValue ("gl_fogblue", fog_blue);

	data = cl.worldmodel->entities;
	if (!data)
		return;

	data = COM_Parse(data);
	if (!data) //should never happen
		return; // error
	if (com_token[0] != '{') //should never happen
		return; // error
	while (1)
	{
		data = COM_Parse(data);
		if (!data)
			return; // error
		if (com_token[0] == '}')
			break; // end of worldspawn
		if (com_token[0] == '_')
			strcpy(key, com_token + 1); // Support "_sky" and "_fog" also
		else
			strcpy(key, com_token);
		while (key[strlen(key)-1] == ' ') // remove trailing spaces
			key[strlen(key)-1] = 0;
		data = COM_Parse(data);
		if (!data)
			return; // error
		strcpy(value, com_token);

		if (!strcmp("sky", key) && value[0])
			R_LoadSkyBox(value);
		//also accept non-standard keys
		else if (!strcmp("skyname", key) && value[0]) //half-life
			R_LoadSkyBox(value);
		else if (!strcmp("qlsky", key) && value[0]) //quake lives
			R_LoadSkyBox(value);
		else if (!strcmp("fog", key) && value[0])
		{
			sscanf(value, "%f %f %f %f", &fog_density, &fog_red, &fog_green, &fog_blue);

			Cvar_SetValue ("gl_fogenable", fog_density ? 1 : 0);
			Cvar_SetValue ("gl_fogdensity", fog_density);
			Cvar_SetValue ("gl_fogred", fog_red);
			Cvar_SetValue ("gl_foggreen", fog_green);
			Cvar_SetValue ("gl_fogblue", fog_blue);
		}
	}
}

/*
=================
R_Sky_f
=================
*/
void R_Sky_f (void)
{
	switch (Cmd_Argc())
	{
		case 1:
			Con_Printf("\"sky\" is \"%s\"\n", skybox_name);
			break;
		case 2:
			R_LoadSkyBox(Cmd_Argv(1));
			break;
		default:
			Con_Printf("usage: sky <skyname>\n");
	}
}

/*
==============
R_SkyEmitSkyBoxVertex
==============
*/
void R_SkyEmitSkyBoxVertex (float s, float t, int axis)
{
	vec3_t		v, b;
	int			j, k;
	float		w, h;

	b[0] = s * FARCLIP / sqrt(3.0);
	b[1] = t * FARCLIP / sqrt(3.0);
	b[2] = FARCLIP / sqrt(3.0);

	for (j=0 ; j<3 ; j++)
	{
		k = st_to_vec[axis][j];
		if (k < 0)
			v[j] = -b[-k - 1];
		else
			v[j] = b[k - 1];
		v[j] += r_origin[j];
	}

	// convert from range [-1,1] to [0,1]
	s = (s+1)*0.5;
	t = (t+1)*0.5;

	// avoid bilerp seam
	w = skyboxtextures[skytexorder[axis]]->width;
	h = skyboxtextures[skytexorder[axis]]->height;

	s = s * (w-1)/w + 0.5/w;
	t = t * (h-1)/h + 0.5/h;

	t = 1.0 - t;

	glTexCoord2f (s, t);
	glVertex3fv (v);
}

/*
==============
R_SkyDrawSkyBox

FIXME: eliminate cracks by adding an extra vert on tjuncs
==============
*/
void R_SkyDrawSkyBox (void)
{
	int		i;

	for (i=0 ; i<6 ; i++)
	{
		if (skymins[0][i] >= skymaxs[0][i] || skymins[1][i] >= skymaxs[1][i])
			continue;

		GL_Bind (skyboxtextures[skytexorder[i]]);

//FIXME: this is to avoid tjunctions until i can do it the right way
		skymins[0][i] = -1;
		skymins[1][i] = -1;
		skymaxs[0][i] = 1;
		skymaxs[1][i] = 1;

		glBegin (GL_QUADS);
		R_SkyEmitSkyBoxVertex (skymins[0][i], skymins[1][i], i);
		R_SkyEmitSkyBoxVertex (skymins[0][i], skymaxs[1][i], i);
		R_SkyEmitSkyBoxVertex (skymaxs[0][i], skymaxs[1][i], i);
		R_SkyEmitSkyBoxVertex (skymaxs[0][i], skymins[1][i], i);
		glEnd ();

		if (R_FogGetDensity() > 0 && r_skyfog.value > 0)
		{
			float *c;

			c = R_FogGetColor();
			glEnable (GL_BLEND);
			glDisable (GL_TEXTURE_2D);
			glColor4f (c[0],c[1],c[2], CLAMP(0.0,r_skyfog.value,1.0));

			glBegin (GL_QUADS);
			R_SkyEmitSkyBoxVertex (skymins[0][i], skymins[1][i], i);
			R_SkyEmitSkyBoxVertex (skymins[0][i], skymaxs[1][i], i);
			R_SkyEmitSkyBoxVertex (skymaxs[0][i], skymaxs[1][i], i);
			R_SkyEmitSkyBoxVertex (skymaxs[0][i], skymins[1][i], i);
			glEnd ();

			glColor3f (1, 1, 1);
			glEnable (GL_TEXTURE_2D);
			glDisable (GL_BLEND);
		}
	}
}

/*
=============================================================

	RENDER CLOUDS

=============================================================
*/
/*
==============
R_SkySetBoxVert
==============
*/
void R_SkySetBoxVert (float s, float t, int axis, vec3_t v)
{
	vec3_t		b;
	int			j, k;

	b[0] = s * FARCLIP / sqrt(3.0);
	b[1] = t * FARCLIP / sqrt(3.0);
	b[2] = FARCLIP / sqrt(3.0);

	for (j=0 ; j<3 ; j++)
	{
		k = st_to_vec[axis][j];
		if (k < 0)
			v[j] = -b[-k - 1];
		else
			v[j] = b[k - 1];
		v[j] += r_origin[j];
	}
}

/*
=============
R_SkyGetTexCoord
=============
*/
void R_SkyGetTexCoord (vec3_t v, float speed, float *s, float *t)
{
	vec3_t	dir;
	float	length, scroll;

	VectorSubtract (v, r_origin, dir);
	dir[2] *= 3;	// flatten the sphere

	length = dir[0]*dir[0] + dir[1]*dir[1] + dir[2]*dir[2];
	length = sqrt (length);
	length = 6*63/length;

	scroll = cl.time*speed;
	scroll -= (int)scroll & ~127;

	*s = (scroll + dir[0] * length) * (1.0/128);
	*t = (scroll + dir[1] * length) * (1.0/128);
}

/*
=============
R_SkyDrawFaceQuad
=============
*/
void R_SkyDrawFaceQuad (glpoly_t *p)
{
	int			i;
	float		*v;
	float		s, t;
	float		skyalpha;

	skyalpha = CLAMP(0.0, r_skyalpha.value, 1.0);
	if (gl_mtexable && skyalpha >= 1.0)
	{
		GL_Bind(solidskytexture);
		GL_EnableMultitexture (); // selects TEXTURE1
		GL_Bind (alphaskytexture);
		glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);

		glBegin (GL_QUADS);
		for (i=0, v=p->verts[0] ; i<4 ; i++, v+=VERTEXSIZE)
		{
			R_SkyGetTexCoord (v, 8, &s, &t);
			qglMultiTexCoord2f (TEXTURE0, s, t);
			R_SkyGetTexCoord (v, 16, &s, &t);
			qglMultiTexCoord2f (TEXTURE1, s, t);
			glVertex3fv (v);
		}
		glEnd ();

		GL_DisableMultitexture();
	}
	else
	{
		GL_Bind(solidskytexture);

		if (skyalpha < 1.0)
			glColor3f (1, 1, 1);

		glBegin (GL_QUADS);
		for (i=0, v=p->verts[0] ; i<4 ; i++, v+=VERTEXSIZE)
		{
			R_SkyGetTexCoord (v, 8, &s, &t);
			glTexCoord2f (s, t);
			glVertex3fv (v);
		}
		glEnd ();
		
		GL_Bind (alphaskytexture);
		glEnable (GL_BLEND);

		if (skyalpha < 1.0)
			glColor4f (1, 1, 1, skyalpha);

		glBegin (GL_QUADS);
		for (i=0, v=p->verts[0] ; i<4 ; i++, v+=VERTEXSIZE)
		{
			R_SkyGetTexCoord (v, 16, &s, &t);
			glTexCoord2f (s, t);
			glVertex3fv (v);
		}
		glEnd ();

		glDisable (GL_BLEND);
	}

	if (R_FogGetDensity() > 0 && r_skyfog.value > 0)
	{
		float *c;

		c = R_FogGetColor();

		glEnable (GL_BLEND);
		glDisable (GL_TEXTURE_2D);
		glColor4f (c[0],c[1],c[2], CLAMP(0.0,r_skyfog.value,1.0));

		glBegin (GL_QUADS);
		for (i=0, v=p->verts[0] ; i<4 ; i++, v+=VERTEXSIZE)
			glVertex3fv (v);
		glEnd ();

		glColor3f (1, 1, 1);
		glEnable (GL_TEXTURE_2D);
		glDisable (GL_BLEND);
	}
}

/*
=============
R_SkyDrawFace
=============
*/
void R_SkyDrawFace (int axis)
{
	glpoly_t	*p;
	vec3_t		verts[4];
	int			i, j, start;
	float		di,qi,dj,qj;
	vec3_t		vup, vright, temp1, temp2;

	R_SkySetBoxVert(-1.0, -1.0, axis, verts[0]);
	R_SkySetBoxVert(-1.0,  1.0, axis, verts[1]);
	R_SkySetBoxVert(1.0,   1.0, axis, verts[2]);
	R_SkySetBoxVert(1.0,  -1.0, axis, verts[3]);

	start = Hunk_LowMark ();
	p = Hunk_AllocName (sizeof(glpoly_t), "skyface");

	VectorSubtract(verts[2],verts[3],vup);
	VectorSubtract(verts[2],verts[1],vright);

	di = CLAMP(4.0, floor(r_skyquality.value), 64.0);
	qi = 1.0 / di;
	dj = (axis < 4) ? di*2 : di; // subdivide vertically more than horizontally on skybox sides
	qj = 1.0 / dj;

	for (i=0; i<di; i++)
	{
		for (j=0; j<dj; j++)
		{
			if (i*qi < skymins[0][axis]/2+0.5 - qi || i*qi > skymaxs[0][axis]/2+0.5 ||
				j*qj < skymins[1][axis]/2+0.5 - qj || j*qj > skymaxs[1][axis]/2+0.5)
				continue;

			//if (i&1 ^ j&1) continue; // checkerboard test
			VectorScale (vright, qi*i, temp1);
			VectorScale (vup, qj*j, temp2);
			VectorAdd(temp1,temp2,temp1);
			VectorAdd(verts[0],temp1,p->verts[0]);

			VectorScale (vup, qj, temp1);
			VectorAdd (p->verts[0],temp1,p->verts[1]);

			VectorScale (vright, qi, temp1);
			VectorAdd (p->verts[1],temp1,p->verts[2]);

			VectorAdd (p->verts[0],temp1,p->verts[3]);

			R_SkyDrawFaceQuad (p);
		}
	}
	Hunk_FreeToLowMark (start);
}

/*
=================
R_SkyProjectPoly

update sky bounds
=================
*/
void R_SkyProjectPoly (int nump, vec3_t vecs)
{
	int		i,j;
	vec3_t	v, av;
	float	s, t, dv;
	int		axis;
	float	*vp;

	// decide which face it maps to
	VectorCopy (vec3_origin, v);
	for (i=0, vp=vecs ; i<nump ; i++, vp+=3)
	{
		VectorAdd (vp, v, v);
	}
	av[0] = fabs(v[0]);
	av[1] = fabs(v[1]);
	av[2] = fabs(v[2]);
	if (av[0] > av[1] && av[0] > av[2])
	{
		if (v[0] < 0)
			axis = 1;
		else
			axis = 0;
	}
	else if (av[1] > av[2] && av[1] > av[0])
	{
		if (v[1] < 0)
			axis = 3;
		else
			axis = 2;
	}
	else
	{
		if (v[2] < 0)
			axis = 5;
		else
			axis = 4;
	}

	// project new texture coords
	for (i=0 ; i<nump ; i++, vecs+=3)
	{
		j = vec_to_st[axis][2];
		if (j > 0)
			dv = vecs[j - 1];
		else
			dv = -vecs[-j - 1];

		j = vec_to_st[axis][0];
		if (j < 0)
			s = -vecs[-j -1] / dv;
		else
			s = vecs[j-1] / dv;
		j = vec_to_st[axis][1];
		if (j < 0)
			t = -vecs[-j -1] / dv;
		else
			t = vecs[j-1] / dv;

		if (s < skymins[0][axis])
			skymins[0][axis] = s;
		if (t < skymins[1][axis])
			skymins[1][axis] = t;
		if (s > skymaxs[0][axis])
			skymaxs[0][axis] = s;
		if (t > skymaxs[1][axis])
			skymaxs[1][axis] = t;
	}
}

/*
=================
R_SkyClipPoly
=================
*/
void R_SkyClipPoly (int nump, vec3_t vecs, int stage)
{
	float	*norm;
	float	*v;
	qboolean	front, back;
	float	d, e;
	float	dists[MAX_CLIP_VERTS];
	int		sides[MAX_CLIP_VERTS];
	vec3_t	newv[2][MAX_CLIP_VERTS];
	int		newc[2];
	int		i, j;

	if (nump > MAX_CLIP_VERTS-2)
		Sys_Error ("R_SkyClipPoly: MAX_CLIP_VERTS");
	if (stage == 6) // fully clipped
	{
		R_SkyProjectPoly (nump, vecs);
		return;
	}

	front = back = false;
	norm = skyclip[stage];
	for (i=0, v = vecs ; i<nump ; i++, v+=3)
	{
		d = DotProduct (v, norm);
		if (d > ON_EPSILON)
		{
			front = true;
			sides[i] = SIDE_FRONT;
		}
		else if (d < ON_EPSILON)
		{
			back = true;
			sides[i] = SIDE_BACK;
		}
		else
			sides[i] = SIDE_ON;
		dists[i] = d;
	}

	if (!front || !back)
	{	// not clipped
		R_SkyClipPoly (nump, vecs, stage+1);
		return;
	}

	// clip it
	sides[i] = sides[0];
	dists[i] = dists[0];
	VectorCopy (vecs, (vecs+(i*3)) );
	newc[0] = newc[1] = 0;

	for (i=0, v = vecs ; i<nump ; i++, v+=3)
	{
		switch (sides[i])
		{
		case SIDE_FRONT:
			VectorCopy (v, newv[0][newc[0]]);
			newc[0]++;
			break;
		case SIDE_BACK:
			VectorCopy (v, newv[1][newc[1]]);
			newc[1]++;
			break;
		case SIDE_ON:
			VectorCopy (v, newv[0][newc[0]]);
			newc[0]++;
			VectorCopy (v, newv[1][newc[1]]);
			newc[1]++;
			break;
		}

		if (sides[i] == SIDE_ON || sides[i+1] == SIDE_ON || sides[i+1] == sides[i])
			continue;

		d = dists[i] / (dists[i] - dists[i+1]);
		for (j=0 ; j<3 ; j++)
		{
			e = v[j] + d*(v[j+3] - v[j]);
			newv[0][newc[0]][j] = e;
			newv[1][newc[1]][j] = e;
		}
		newc[0]++;
		newc[1]++;
	}

	// continue
	R_SkyClipPoly (newc[0], newv[0][0], stage+1);
	R_SkyClipPoly (newc[1], newv[1][0], stage+1);
}

/*
================
R_SkyProcessPoly
================
*/
void R_SkyProcessPoly (glpoly_t *p)
{
	int			i;
	vec3_t		verts[MAX_CLIP_VERTS];

	//draw it (just make it transparent)
	R_DrawGLPoly34 (p);

	//update sky bounds
	if (!r_fastsky.value)
	{
		for (i=0 ; i<p->numverts ; i++)
			VectorSubtract (p->verts[i], r_origin, verts[i]);
		R_SkyClipPoly (p->numverts, verts[0], 0);
	}
}

/*
================
R_SkyProcessTextureChains

handles sky polys in world model
================
*/
void R_SkyProcessTextureChains (void)
{
	int			i;
	msurface_t	*s;
	texture_t	*t;

	if (!r_drawworld.value)
		return;

	for (i=0 ; i<cl.worldmodel->numtextures ; i++)
	{
		t = cl.worldmodel->textures[i];

		if (!t || !t->texturechain || !(t->texturechain->flags & SURF_DRAWSKY))
			continue;

		for (s = t->texturechain; s; s = s->texturechain)
			if (!s->culled)
				R_SkyProcessPoly (s->polys);
	}
}

/*
================
R_SkyProcessEntities

handles sky polys on brush models
================
*/
void R_SkyProcessEntities (void)
{
	entity_t	*e;
	msurface_t	*s;
	glpoly_t	*p;
	int			i,j,k,mark;
	float		dot;
	qboolean	rotated;
	vec3_t		temp, forward, right, up;

	if (!r_drawentities.value)
		return;

	for (i=0 ; i<cl_numvisedicts ; i++)
	{
		e = cl_visedicts[i];

		if (e->model->type != mod_brush)
			continue;

		if (R_CullModelForEntity(e))
			continue;

		if (e->alpha == ENTALPHA_ZERO)
			continue;

		VectorSubtract (r_refdef.vieworg, e->origin, modelorg);
		if (e->angles[0] || e->angles[1] || e->angles[2])
		{
			rotated = true;
			AngleVectors (e->angles, forward, right, up);
			VectorCopy (modelorg, temp);
			modelorg[0] = DotProduct (temp, forward);
			modelorg[1] = -DotProduct (temp, right);
			modelorg[2] = DotProduct (temp, up);
		}
		else
			rotated = false;

		s = &e->model->surfaces[e->model->firstmodelsurface];

		for (j=0 ; j<e->model->nummodelsurfaces ; j++, s++)
		{
			if (s->flags & SURF_DRAWSKY)
			{
				dot = DotProduct (modelorg, s->plane->normal) - s->plane->dist;
				if (((s->flags & SURF_PLANEBACK) && (dot < -BACKFACE_EPSILON)) ||
					(!(s->flags & SURF_PLANEBACK) && (dot > BACKFACE_EPSILON)))
				{
					//copy the polygon and translate manually, since R_SkyProcessPoly needs it to be in world space
					mark = Hunk_LowMark();
					p = Hunk_AllocName (sizeof(*s->polys), "skypoly"); //FIXME: don't allocate for each poly
					p->numverts = s->polys->numverts;
					for (k=0; k<p->numverts; k++)
					{
						if (rotated)
						{
							p->verts[k][0] = e->origin[0] + s->polys->verts[k][0] * forward[0]
														  - s->polys->verts[k][1] * right[0]
														  + s->polys->verts[k][2] * up[0];
							p->verts[k][1] = e->origin[1] + s->polys->verts[k][0] * forward[1]
														  - s->polys->verts[k][1] * right[1]
														  + s->polys->verts[k][2] * up[1];
							p->verts[k][2] = e->origin[2] + s->polys->verts[k][0] * forward[2]
														  - s->polys->verts[k][1] * right[2]
														  + s->polys->verts[k][2] * up[2];
						}
						else
							VectorAdd(s->polys->verts[k], e->origin, p->verts[k]);
					}
					R_SkyProcessPoly (p);
					Hunk_FreeToLowMark (mark);
				}
			}
		}
	}
}

/*
==============
R_SkyDrawSkyLayers

draws the old-style scrolling cloud layers
==============
*/
void R_SkyDrawSkyLayers (void)
{
	int i;
	float	skyalpha;

	skyalpha = CLAMP(0.0, r_skyalpha.value, 1.0);
	if (skyalpha < 1.0)
		glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	for (i=0 ; i<6 ; i++)
		if (skymins[0][i] < skymaxs[0][i] && skymins[1][i] < skymaxs[1][i])
			R_SkyDrawFace (i); // draw skybox arround the world

	if (skyalpha < 1.0)
		glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
}

/*
==============
R_DrawSky

called once per frame before drawing anything else
==============
*/
void R_DrawSky (void)
{
	int				i;

	//
	// reset sky bounds
	//
	for (i=0 ; i<6 ; i++)
	{
		skymins[0][i] = skymins[1][i] = 9999;
		skymaxs[0][i] = skymaxs[1][i] = -9999;
	}

	//
	// process world and bmodels: draw flat-shaded sky surfs, and update skybounds
	//
	R_FogDisableGFog ();

	glDisable (GL_TEXTURE_2D);
	if (R_FogGetDensity() > 0)
		glColor3fv (R_FogGetColor());
	else
		glColor3fv (skyflatcolor);

	R_SkyProcessTextureChains ();
	R_SkyProcessEntities ();

	glColor3f (1, 1, 1);
	glEnable (GL_TEXTURE_2D);

	//
	// render slow sky: cloud layers or skybox
	//
	if (!r_fastsky.value && !(R_FogGetDensity() > 0 && r_skyfog.value >= 1))
	{
		glDepthFunc(GL_GEQUAL);
		glDepthMask (GL_FALSE); // don't bother writing Z

		if (!r_oldsky.value && !oldsky)
			R_SkyDrawSkyBox ();
		else
			R_SkyDrawSkyLayers();

		glDepthMask (GL_TRUE); // back to normal Z buffering
		glDepthFunc(GL_LEQUAL);
	}

	R_FogEnableGFog ();
}

/*
=============
R_InitSky

A sky texture is 256*128, with the right side being a masked overlay
==============
*/
void R_InitSky (texture_t *mt)
{
	char		texturename[64];
	int			i, j, p, r, g, b, count;
	int			scaledx;
	byte		*src;
	byte		fixedsky[256*128];
	static byte	front_data[128*128];
	static byte	back_data[128*128];
	unsigned	*rgba;

	src = (byte *)mt + mt->offsets[0];

	if (mt->width * mt->height != sizeof(fixedsky))
	{
		Con_DPrintf ("R_InitSky: non-standard sky texture '%s' (%dx%d, should be 256x128)\n", mt->name, mt->width, mt->height);

		// Resize sky texture to correct size
		memset (fixedsky, 0, sizeof(fixedsky));

		for (i = 0; i < 256; ++i)
		{
			scaledx = i * mt->width / 256 * mt->height;

			for (j = 0; j < 128; ++j)
				fixedsky[i * 128 + j] = src[scaledx + j * mt->height / 128];
		}

		src = fixedsky;
	}

// extract back layer and upload
	for (i=0 ; i<128 ; i++)
	{
		for (j=0 ; j<128 ; j++)
		{
			back_data[(i*128) + j] = src[i*256 + j + 128];
		}
	}

	sprintf (texturename, "%s:%s_back", loadmodel->name, mt->name);
	solidskytexture = GL_LoadTexture (loadmodel, texturename, 128, 128, SRC_INDEXED, back_data, TEXPREF_NONE);

// extract front layer and upload
	for (i=0 ; i<128 ; i++)
	{
		for (j=0 ; j<128 ; j++)
		{
			front_data[(i*128) + j] = src[i*256 + j];
			if (front_data[(i*128) + j] == 0)
				front_data[(i*128) + j] = 255;
		}
	}

	sprintf (texturename, "%s:%s_front", loadmodel->name, mt->name);
	alphaskytexture = GL_LoadTexture (loadmodel, texturename, 128, 128, SRC_INDEXED, front_data, TEXPREF_ALPHA);

// calculate r_fastsky color based on average of all opaque foreground colors
	r = g = b = count = 0;
	for (i=0 ; i<128 ; i++)
	{
		for (j=0 ; j<128 ; j++)
		{
			p = src[i*256 + j];
			if (p != 0)
			{
				rgba = &d_8to24table[p];
				r += ((byte *)rgba)[0];
				g += ((byte *)rgba)[1];
				b += ((byte *)rgba)[2];
				count++;
			}
		}
	}

	skyflatcolor[0] = (float)r/(count*255);
	skyflatcolor[1] = (float)g/(count*255);
	skyflatcolor[2] = (float)b/(count*255);
}

//==================================================================================================================

/*
============================================================================== 

	LIGHT BLOOMS ROUTINES

============================================================================== 
*/

static float Diamond8x[8][8] = 
{
	{0.0f, 0.0f, 0.0f, 0.1f, 0.1f, 0.0f, 0.0f, 0.0f},
	{0.0f, 0.0f, 0.2f, 0.3f, 0.3f, 0.2f, 0.0f, 0.0f},
	{0.0f, 0.2f, 0.4f, 0.6f, 0.6f, 0.4f, 0.2f, 0.0f},
	{0.1f, 0.3f, 0.6f, 0.9f, 0.9f, 0.6f, 0.3f, 0.1f},
	{0.1f, 0.3f, 0.6f, 0.9f, 0.9f, 0.6f, 0.3f, 0.1f},
	{0.0f, 0.2f, 0.4f, 0.6f, 0.6f, 0.4f, 0.2f, 0.0f},
	{0.0f, 0.0f, 0.2f, 0.3f, 0.3f, 0.2f, 0.0f, 0.0f},
	{0.0f, 0.0f, 0.0f, 0.1f, 0.1f, 0.0f, 0.0f, 0.0f}
};

static float Diamond6x[6][6] = 
{
	{0.0f, 0.0f, 0.1f, 0.1f, 0.0f, 0.0f},
	{0.0f, 0.3f, 0.5f, 0.5f, 0.3f, 0.0f},
	{0.1f, 0.5f, 0.9f, 0.9f, 0.5f, 0.1f},
	{0.1f, 0.5f, 0.9f, 0.9f, 0.5f, 0.1f},
	{0.0f, 0.3f, 0.5f, 0.5f, 0.3f, 0.0f},
	{0.0f, 0.0f, 0.1f, 0.1f, 0.0f, 0.0f}
};

static float Diamond4x[4][4] = 
{
	{0.3f, 0.4f, 0.4f, 0.3f},
	{0.4f, 0.9f, 0.9f, 0.4f},
	{0.4f, 0.9f, 0.9f, 0.4f},
	{0.3f, 0.4f, 0.4f, 0.3f}
};

static int		bloomsize;
static float	bloomalpha;
static float	bloomintensity;

cvar_t	r_bloom_enable			= {"r_bloom_enable", "0", true};
cvar_t	r_bloom_darken			= {"r_bloom_darken", "2", true};
cvar_t	r_bloom_alpha			= {"r_bloom_alpha", "0.2", true};
cvar_t	r_bloom_intensity		= {"r_bloom_intensity", "0.8", true};
cvar_t	r_bloom_diamond_size	= {"r_bloom_diamond_size", "8", true};
cvar_t	r_bloom_sample_size		= {"r_bloom_sample_size", "512", true};
cvar_t	r_bloom_fast_sample		= {"r_bloom_fast_sample", "0", true};

gltexture_t		*bloomscreentexture;
gltexture_t		*bloomeffecttexture;
gltexture_t		*bloombackuptexture;
gltexture_t		*bloomdownsamplingtexture;

static int	screendownsamplingtexture_size;
static int	screen_texture_width, screen_texture_height;
static int	screenbackuptexture_width, screenbackuptexture_height;

// Current refdef size:
static int	curView_x;
static int	curView_y;
static int	curView_width;
static int	curView_height;

// Texture coordinates of screen data inside screentexture.
static float	screenText_tcw;
static float	screenText_tch;

static int	sample_width;
static int	sample_height;

// Texture coordinates of adjusted textures.
static float	sampleText_tcw;
static float	sampleText_tch;

// This macro is in sample size workspace coordinates.
#define R_BloomSamplePass( xpos, ypos )                           \
    glBegin(GL_QUADS);                                             \
    glTexCoord2f(  0,                      sampleText_tch);        \
    glVertex2f(    xpos,                   ypos);                  \
    glTexCoord2f(  0,                      0);                     \
    glVertex2f(    xpos,                   ypos+sample_height);    \
    glTexCoord2f(  sampleText_tcw,         0);                     \
    glVertex2f(    xpos+sample_width,      ypos+sample_height);    \
    glTexCoord2f(  sampleText_tcw,         sampleText_tch);        \
    glVertex2f(    xpos+sample_width,      ypos);                  \
    glEnd();

#define R_BloomQuad( x, y, width, height, textwidth, textheight ) \
    glBegin(GL_QUADS);                                             \
    glTexCoord2f(  0,          textheight);                        \
    glVertex2f(    x,          y);                                 \
    glTexCoord2f(  0,          0);                                 \
    glVertex2f(    x,          y+height);                          \
    glTexCoord2f(  textwidth,  0);                                 \
    glVertex2f(    x+width,    y+height);                          \
    glTexCoord2f(  textwidth,  textheight);                        \
    glVertex2f(    x+width,    y);                                 \
    glEnd();

/*
=================
R_BloomInitBackUpTexture
=================
*/
void R_BloomInitBackUpTexture (int width, int height)
{
	byte   *data;

	data = malloc (width * height * sizeof(int));
	memset (data, 0, width * height * sizeof(int));

	screenbackuptexture_width  = width;
	screenbackuptexture_height = height;

	bloombackuptexture = GL_LoadTexture (NULL, "bloombackuptexture", width, height, SRC_BLOOM, data, TEXPREF_LUMA);

	free (data);
}

/*
=================
R_BloomInitEffectTexture
=================
*/
void R_BloomInitEffectTexture (void)
{
	byte   *data;
	float   bloomsizecheck;

	if (r_bloom_sample_size.value < 32)
		Cvar_SetValue ("r_bloom_sample_size", 32);

	// Make sure bloom size is a power of 2.
	bloomsize = r_bloom_sample_size.value;
	bloomsizecheck = (float)bloomsize;

	while(bloomsizecheck > 1.0f)
		bloomsizecheck /= 2.0f;

	if (bloomsizecheck != 1.0f)
	{
		bloomsize = 32;
		while (bloomsize < r_bloom_sample_size.value)
			bloomsize *= 2;
	}

	// Make sure bloom size doesn't have stupid values.
	if (bloomsize > screen_texture_width || bloomsize > screen_texture_height)
		bloomsize = min( screen_texture_width, screen_texture_height );

	if (bloomsize != r_bloom_sample_size.value)
		Cvar_SetValue ("r_bloom_sample_size", bloomsize);

	data = malloc (bloomsize * bloomsize * sizeof(int));
	memset (data, 0, bloomsize * bloomsize * sizeof(int));

	bloomeffecttexture = GL_LoadTexture (NULL, "bloomeffecttexture", bloomsize, bloomsize, SRC_BLOOM, data, TEXPREF_LUMA);

	free (data);
}

/*
=================
R_BloomInitTextures
=================
*/
void R_BloomInitTextures (void)
{
	byte   *data;
	int     size;

	// Find closer power of 2 to screen size.
	for (screen_texture_width = 2; screen_texture_width < glwidth; screen_texture_width *= 2)
		;
	for (screen_texture_height = 2; screen_texture_height < glheight; screen_texture_height *= 2)
		;

	// Disable blooms if we can't handle a texture of that size.
	if (screen_texture_width > gl_texture_max_size || screen_texture_height > gl_texture_max_size)
	{
		screen_texture_width = screen_texture_height = 0;
		Cvar_SetValue ("r_bloom_enable", 0);
		Con_Warning ("R_BloomInitTextures: too high resolution for Light Bloom. Effect disabled\n");
		return;
	}

	// Init the screen texture.
	size = screen_texture_width * screen_texture_height * sizeof(int);
	data = malloc (size);
	memset (data, 255, size);

	bloomscreentexture = GL_LoadTexture (NULL, "bloomscreentexture", screen_texture_width, screen_texture_height, SRC_BLOOM, data, TEXPREF_LUMA);

	free (data);

	// Validate bloom size and init the bloom effect texture.
	R_BloomInitEffectTexture ();

	// If screensize is more than 2x the bloom effect texture, set up for stepped downsampling.
	bloomdownsamplingtexture = NULL;
	screendownsamplingtexture_size = 0;
	if (glwidth > (bloomsize * 2) && !r_bloom_fast_sample.value)
	{
		screendownsamplingtexture_size = (int)(bloomsize * 2);
		data = malloc (screendownsamplingtexture_size * screendownsamplingtexture_size * sizeof(int));
		memset (data, 0, screendownsamplingtexture_size * screendownsamplingtexture_size * sizeof(int));

		bloomdownsamplingtexture = GL_LoadTexture (NULL, "bloomdownsamplingtexture", screendownsamplingtexture_size, screendownsamplingtexture_size, SRC_BLOOM, data, TEXPREF_LUMA);

		free (data);
	}

	// Init the screen backup texture.
	if (screendownsamplingtexture_size)
		R_BloomInitBackUpTexture (screendownsamplingtexture_size, screendownsamplingtexture_size);
	else
		R_BloomInitBackUpTexture (bloomsize, bloomsize);
}

/*
=================
R_InitBloomTextures
=================
*/
void R_InitBloomTextures (void)
{
	bloomsize = 0;

	if (!r_bloom_enable.value)
		return;

	R_BloomInitTextures ();
}

/*
=================
R_BloomDrawEffect
=================
*/
void R_BloomDrawEffect (void)
{
	GL_Bind (bloomeffecttexture);

	glEnable(GL_BLEND);
	glBlendFunc(GL_ONE, GL_ONE);

	glColor4f(bloomalpha, bloomalpha, bloomalpha, 1.0f);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	glBegin(GL_QUADS);

	glTexCoord2f(0, sampleText_tch);
	glVertex2f(curView_x, curView_y);

	glTexCoord2f(0, 0);
	glVertex2f(curView_x, curView_y + curView_height);

	glTexCoord2f(sampleText_tcw, 0);
	glVertex2f(curView_x + curView_width, curView_y + curView_height);

	glTexCoord2f(sampleText_tcw, sampleText_tch);
	glVertex2f(curView_x + curView_width, curView_y);

	glEnd();

	glDisable(GL_BLEND);
}

/*
=================
R_BloomGeneratexDiamonds
=================
*/
void R_BloomGeneratexDiamonds (void)
{
	int         i, j;
	static float intensity;

	// Setup sample size workspace
	glViewport(0, 0, sample_width, sample_height);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity ();
	glOrtho(0, sample_width, sample_height, 0, -10, 100);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity ();

	// Copy small scene into bloomeffecttexture.
	GL_Bind (bloomeffecttexture);
	glCopyTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, 0, 0, sample_width, sample_height);

	// Start modifying the small scene corner.
	glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
	glEnable(GL_BLEND);

	// Darkening passes
	if (r_bloom_darken.value)
	{
		glBlendFunc(GL_DST_COLOR, GL_ZERO);
		glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

		for (i = 0; i < r_bloom_darken.value; i++) 
		{
			R_BloomSamplePass (0, 0);
		}
		glCopyTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, 0, 0, sample_width, sample_height);
	}

	// Bluring passes.
	glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_COLOR);

	if (r_bloom_diamond_size.value > 7 || r_bloom_diamond_size.value <= 3)
	{
		if (r_bloom_diamond_size.value != 8)
			Cvar_SetValue("r_bloom_diamond_size", 8);

		for (i = 0; i < r_bloom_diamond_size.value; i++) 
		{
			for (j = 0; j < r_bloom_diamond_size.value; j++) 
			{
				intensity = bloomintensity * 0.3 * Diamond8x[i][j];
				if (intensity < 0.01f)
					continue;

				glColor4f(intensity, intensity, intensity, 1.0);
				R_BloomSamplePass (i - 4, j - 4);
			}
		}
	}
	else if (r_bloom_diamond_size.value > 5)
	{
		if (r_bloom_diamond_size.value != 6) 
			Cvar_SetValue("r_bloom_diamond_size", 6);

		for (i = 0; i < r_bloom_diamond_size.value; i++) 
		{
			for (j = 0; j < r_bloom_diamond_size.value; j++) 
			{
				intensity = bloomintensity * 0.5 * Diamond6x[i][j];
				if (intensity < 0.01f)
					continue;

				glColor4f(intensity, intensity, intensity, 1.0);
				R_BloomSamplePass(i - 3, j - 3);
			}
		}
	}
	else if (r_bloom_diamond_size.value > 3)
	{
		if (r_bloom_diamond_size.value != 4)
			Cvar_SetValue("r_bloom_diamond_size", 4);

		for (i = 0; i < r_bloom_diamond_size.value; i++)
		{
			for (j = 0; j < r_bloom_diamond_size.value; j++)
			{
				intensity = bloomintensity * 0.8f * Diamond4x[i][j];
				if (intensity < 0.01f)
					continue;

				glColor4f(intensity, intensity, intensity, 1.0);
				R_BloomSamplePass (i - 2, j - 2);
			}
		}
	}

	glCopyTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, 0, 0, sample_width, sample_height);

	// Restore full screen workspace.
	glViewport(0, 0, glwidth, glheight);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity ();
	glOrtho(0, glwidth, glheight, 0, -10, 100);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity ();
}

/*
=================
R_BloomDownsampleView
=================
*/
void R_BloomDownsampleView (void)
{
	glDisable(GL_BLEND);
	glColor4f(1.0f, 1.0f, 1.0f, 1.0f);

	// Stepped downsample.
	if (screendownsamplingtexture_size)
	{
		int midsample_width = screendownsamplingtexture_size * sampleText_tcw;
		int midsample_height = screendownsamplingtexture_size * sampleText_tch;

		// Copy the screen and draw resized.
		GL_Bind (bloomscreentexture);
		glCopyTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, curView_x, glheight - (curView_y + curView_height), curView_width, curView_height);
		R_BloomQuad (0, glheight - midsample_height, midsample_width, midsample_height, screenText_tcw, screenText_tch);

		// Now copy into Downsampling (mid-sized) texture.
		GL_Bind (bloomdownsamplingtexture);
		glCopyTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, 0, 0, midsample_width, midsample_height);

		// Now draw again in bloom size.
		glColor4f(0.5f, 0.5f, 0.5f, 1.0f);
		R_BloomQuad (0, glheight - sample_height, sample_width, sample_height, sampleText_tcw, sampleText_tch);

		// Now blend the big screen texture into the bloom generation space (hoping it adds some blur).
		glEnable(GL_BLEND);
		glBlendFunc(GL_ONE, GL_ONE);
		glColor4f(0.5f, 0.5f, 0.5f, 1.0f);
		GL_Bind (bloomscreentexture);
		R_BloomQuad (0, glheight - sample_height, sample_width, sample_height, screenText_tcw, screenText_tch);
		glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
		glDisable(GL_BLEND);
	} 
	else
	{
		// Downsample simple.
		GL_Bind (bloomscreentexture);
		glCopyTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, curView_x, glheight - (curView_y + curView_height), curView_width, curView_height);
		R_BloomQuad (0, glheight - sample_height, sample_width, sample_height, screenText_tcw, screenText_tch);
	}
}

/*
=================
R_BloomBlend
=================
*/
void R_BloomBlend (void)
{
	if (!r_bloom_enable.value)
		return;

	if (!bloomsize || screen_texture_width < glwidth || screen_texture_height < glheight)
		R_BloomInitTextures();

	if (screen_texture_width < bloomsize || screen_texture_height < bloomsize)
		return;

	// Set up full screen workspace.
	glViewport(0, 0, glwidth, glheight);
	glDisable(GL_DEPTH_TEST);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity ();
	glOrtho(0, glwidth, glheight, 0, -10, 100);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity ();
	glDisable(GL_CULL_FACE);

	glDisable(GL_BLEND);
	glEnable(GL_TEXTURE_2D);

	glColor4f(1, 1, 1, 1);

	// Setup current sizes
	curView_x = 0;
	curView_y = 0;
	curView_width = glwidth;
	curView_height = glheight;
	screenText_tcw = ((float)glwidth / (float)screen_texture_width);
	screenText_tch = ((float)glheight / (float)screen_texture_height);

	if (glheight > glwidth)
	{
		sampleText_tcw = ((float)glwidth / (float)glheight);
		sampleText_tch = 1.0f;
	}
	else
	{
		sampleText_tcw = 1.0f;
		sampleText_tch = ((float)glheight / (float)glwidth);
	}

	sample_width = bloomsize * sampleText_tcw;
	sample_height = bloomsize * sampleText_tch;

	// Copy the screen space we'll use to work into the backup texture.
	GL_Bind (bloombackuptexture);
	glCopyTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, 0, 0, screenbackuptexture_width, screenbackuptexture_height);

	bloomalpha = CLAMP(0.0, r_bloom_alpha.value, 1.0);
	bloomintensity = r_bloom_intensity.value;

	// Create the bloom image.
	R_BloomDownsampleView();
	R_BloomGeneratexDiamonds();

	// Restore the screen-backup to the screen.
	glDisable(GL_BLEND);
	GL_Bind (bloombackuptexture);
	glColor4f(1, 1, 1, 1);
	R_BloomQuad (0, glheight - screenbackuptexture_height, screenbackuptexture_width, screenbackuptexture_height, 1.0, 1.0);

	R_BloomDrawEffect ();

	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
}

